LIBRARY		scheme16
DESCRIPTION	"MIT/GNU Scheme Win16 Server for Win32S DLL"

EXETYPE		WINDOWS

STUB		"WINSTUB.EXE"

CODE		PRELOAD SHARED FIXED
DATA		PRELOAD SHARED FIXED SINGLE

HEAPSIZE	0

EXPORTS		ntw16lib_init		@ 1
		ntw16lib_handler	@ 2
