#| -*-Scheme-*-

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
    2017, 2018, 2019 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,
USA.

|#

;;;; Tests for library loader

(declare (usual-integrations))

(include "test-library-data/support-code.scm")

(define-test 'read-r7rs-source:r7rs-example
  (lambda ()
    (let ((source (read-r7rs-source r7rs-example-filename))
	  (db (make-library-db 'test)))
      (add-standard-libraries! db)
      (let ((program (register-r7rs-source! source db)))
	(assert-equal (unsyntax (library-contents program))
		      '(begin
			 (define grid (make-grid 24 24))
			 (grid-set! grid 1 1 #t)
			 (grid-set! grid 2 2 #t)
			 (grid-set! grid 3 0 #t)
			 (grid-set! grid 3 1 #t)
			 (grid-set! grid 3 2 #t)
			 (life grid 80)))
	(let ((env (library-imports-environment program)))
	  (assert-lset= eq?
			(environment-bound-names env)
			(map library-import-to
			     (library-imports program))))))))