#include <winver.h>

1 VERSIONINFO
	FILEOS		VOS__WINDOWS32
	FILETYPE	VFT_DLL
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904E4"
    BEGIN
	VALUE "CompanyName", "GNU Project"
	VALUE "FileDescription", "DIB utilities for Scheme Graphics"
	VALUE "FileVersion", "1.0"
	VALUE "InternalName", "dibutils"
	VALUE "LegalCopyright", "Copyright Massachusetts Institute of Technology 1993"
	VALUE "OriginalFilename", "DIBUTILS.DLL"
	VALUE "ProductName", "MIT/GNU Scheme"
	VALUE "ProductVersion", "7.7.1"
    END
  END

  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x0409, 1251
  END
END
